""" Utilities for working with SED documents

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..log.data_model import Status
from ..report.data_model import VariableResults, DataGeneratorResults  # noqa: F401
from ..utils.core import pad_arrays_to_consistent_shapes
from ..warnings import warn
from ..xml.utils import eval_xpath
from .data_model import (SedBase, SedIdGroupMixin, SedDocument,  # noqa: F401
                         Model, ModelLanguagePattern, ModelChange, ModelAttributeChange, AddElementModelChange,
                         ReplaceElementModelChange, RemoveElementModelChange, ComputeModelChange, SetValueComputeModelChange,
                         OneStepSimulation, SteadyStateSimulation, UniformTimeCourseSimulation,
                         Task, RepeatedTask, Output, Report, Plot, Plot2D, Plot3D,
                         DataGenerator, Variable,
                         Range, UniformRange, VectorRange, FunctionalRange, UniformRangeType)
from .math import AGGREGATE_MATH_FUNCTIONS, compile_math, eval_math
from .warnings import InconsistentVariableShapesWarning
from lxml import etree
import copy
import io
import libsedml  # noqa: F401
import numpy
import os
import re
import requests
import tempfile

__all__ = [
    'append_all_nested_children_to_doc',
    'add_namespaces_to_xml_node',
    'convert_xml_node_to_string',
    'get_data_generators_for_output',
    'get_variables_for_data_generators',
    'get_variables_for_task',
    'get_model_changes_for_task',
    'resolve_model_and_apply_xml_changes',
    'resolve_model',
    'apply_changes_to_xml_model',
    'get_values_of_variable_model_xml_targets_of_model_change',
    'get_value_of_variable_model_xml_targets',
    'calc_compute_model_change_new_value',
    'calc_data_generator_results',
    'calc_data_generators_results',
    'remove_model_changes',
    'remove_algorithm_parameter_changes',
    'replace_complex_data_generators_with_generators_for_individual_variables',
    'remove_plots',
    'get_first_last_models_executed_by_task',
    'get_models_referenced_by_task',
    'get_models_referenced_by_range',
    'get_models_referenced_by_model_change',
    'get_range_len',
    'resolve_range',
    'get_namespaces_for_sed_object',
    'get_xml_node_namespace_tag_target',
    'is_model_language_encoded_in_xml',
    'does_model_language_use_xpath_variable_targets',
    'get_task_results_shape',
]


def append_all_nested_children_to_doc(doc):
    """ Append all nested children to a SED document

    Args:
        doc (:obj:`SedDocument`): SED document
    """
    data_generators = set(doc.data_generators)
    tasks = set(doc.tasks)
    simulations = set(doc.simulations)
    models = set(doc.models)

    for model in doc.models:
        for change in model.changes:
            if isinstance(change, ComputeModelChange):
                for var in change.variables:
                    # tasks not added because `var.task` should be null
                    if var.model:
                        models.add(var.model)

    for output in doc.outputs:
        if isinstance(output, Report):
            for data_set in output.data_sets:
                if data_set.data_generator:
                    data_generators.add(data_set.data_generator)

        elif isinstance(output, Plot2D):
            for curve in output.curves:
                if curve.x_data_generator:
                    data_generators.add(curve.x_data_generator)
                if curve.y_data_generator:
                    data_generators.add(curve.y_data_generator)

        elif isinstance(output, Plot3D):
            for surface in output.surfaces:
                if surface.x_data_generator:
                    data_generators.add(surface.x_data_generator)
                if surface.y_data_generator:
                    data_generators.add(surface.y_data_generator)
                if surface.z_data_generator:
                    data_generators.add(surface.z_data_generator)

    for data_gen in data_generators:
        for var in data_gen.variables:
            if var.task:
                tasks.add(var.task)
            # models not added because `var.model` should be null

    for task in tasks:
        if isinstance(task, Task):
            if task.model:
                models.add(task.model)
            if task.simulation:
                simulations.add(task.simulation)

    doc.models += list(models - set(doc.models))
    doc.simulations += list(simulations - set(doc.simulations))
    doc.tasks += list(tasks - set(doc.tasks))
    doc.data_generators += list(data_generators - set(doc.data_generators))


def add_namespaces_to_xml_node(node, namespace_prefixes, namespaces):
    """ Add namespaces to an XML node

    Args:
        node (:obj:`libsedml.XMLNode`): XML node
        namespace_prefixes (:obj:`set` of :obj:`str`): namespace prefixes used in the node which aren't defined in the node
        namespaces (:obj:`libsedml.XMLNamespaces`): namespaces for the parent document
    """
    for i_ns in range(namespaces.getNumNamespaces()):
        uri = namespaces.getURI(i_ns)
        prefix = namespaces.getPrefix(i_ns)

        if prefix in namespace_prefixes:
            node.addNamespace(uri, prefix)


def convert_xml_node_to_string(node):
    """ Generate a string representation of an XML node

    Args:
        node (:obj:`libsedml.XMLNode`): XML node

    Returns:
        :obj:`str`: string representation of the node
    """
    return node.convertXMLNodeToString(node)


def get_data_generators_for_output(output):
    """ Get the data generators involved in an output

    Args:
        output (:obj:`Output`): report or plot

    Returns:
        :obj:`set` of :obj:`DataGenerator`: data generators involved in the output
    """
    data_generators = set()

    if isinstance(output, Report):
        for data_set in output.data_sets:
            data_generators.add(data_set.data_generator)

    elif isinstance(output, Plot2D):
        for curve in output.curves:
            data_generators.add(curve.x_data_generator)
            data_generators.add(curve.y_data_generator)

    elif isinstance(output, Plot3D):
        for surface in output.surfaces:
            data_generators.add(surface.x_data_generator)
            data_generators.add(surface.y_data_generator)
            data_generators.add(surface.z_data_generator)

    else:
        raise NotImplementedError('Output of type {} is not supported.'.format(output.__class__.__name__))

    if None in data_generators:
        data_generators.remove(None)

    return data_generators


def get_variables_for_data_generators(data_generators):
    """ Get the variables involved in a collection of generators

    Args:
        data_generators (:obj:`list` of :obj:`DataGenerator`): data generators

    Returns:
        :obj:`set` of :obj:`Variable`: variables involved in the data generators
    """
    variables = set()
    for data_gen in data_generators:
        variables.update(set(data_gen.variables))
    return variables


def get_variables_for_task(doc, task):
    """ Get the variables that a task must record

    Args:
        doc (:obj:`SedDocument`): SED document
        task (:obj:`Task`): task

    Returns:
        :obj:`list` of :obj:`Variable`: variables that task must record
    """
    data_generators = set()
    for output in doc.outputs:
        data_generators.update(get_data_generators_for_output(output))

    variables = get_variables_for_data_generators(data_generators)

    return list(filter(lambda var: var.task == task, variables))


def get_model_changes_for_task(task):
    """ Get the changes to models for a task

    Args:
        task (:obj:`Task`): task

    Returns:
        :obj:`list` of :obj:`ModelChange`: changes to the model
    """
    if isinstance(task, Task):
        return []

    elif isinstance(task, RepeatedTask):
        changes = list(task.changes)

        for sub_task in task.sub_tasks:
            changes.extend(get_model_changes_for_task(sub_task.task))

        return changes

    else:
        raise NotImplementedError('Task of type {} is not supported.'.format(task.__class__.__name__))


BIOMODELS_DOWNLOAD_ENDPOINT = 'https://www.ebi.ac.uk/biomodels/model/download/{}?filename={}_url.xml'


def resolve_model_and_apply_xml_changes(model, sed_doc, working_dir,
                                        apply_xml_model_changes=True,
                                        save_to_file=True,
                                        pretty_print_modified_xml_models=False):
    """ Resolve the source of a model and, optionally, apply XML changes to the model.

    Args:
        model (:obj:`Model`): model whose ``source`` is one of the following

            * A path to a file
            * A URL
            * A MIRIAM URN for an entry in the BioModelsl database (e.g., ``urn:miriam:biomodels.db:BIOMD0000000012``)
            * A reference to another model, using the ``id`` of the other model (e.g., ``#other-model-id``).
              In this case, the model also inherits changes from the parent model.

        sed_doc (:obj:`SedDocument`): parent SED document; used to resolve sources defined by reference to other models
        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)
        apply_xml_model_changes (:obj:`bool`, optional): if :obj:`True`, apply any model changes specified in the SED-ML file before
            calling :obj:`task_executer`.
        save_to_file (:obj:`bool`): whether to save the resolved/modified model to a file
        pretty_print_modified_xml_models (:obj:`bool`, optional): if :obj:`True`, pretty print modified XML models

    Returns:
        :obj:`tuple`:

            * :obj:`Model`: modified model
            * :obj:`str`: temporary path to the source of the modified model, if the model needed to be resolved from
              a remote source of modified
            * :obj:`etree._Element`: element tree for the resolved/modified model
    """
    model = copy.deepcopy(model)

    # resolve model
    temp_model_source = resolve_model(model, sed_doc, working_dir)

    # apply changes to model
    if apply_xml_model_changes and model.language and is_model_language_encoded_in_xml(model.language):
        # read model from file
        try:
            model_etree = etree.parse(model.source)
        except Exception as exception:
            raise ValueError('The model could not be parsed because the model is not a valid XML document: {}'.format(str(exception)))

        if model.changes:
            # apply changes
            apply_changes_to_xml_model(model, model_etree, sed_doc, working_dir)
            model.changes.clear()

            # write model to file
            if save_to_file:
                if temp_model_source is None:
                    modified_model_file, temp_model_source = tempfile.mkstemp(suffix='.xml')
                    os.close(modified_model_file)
                    model.source = temp_model_source

                model_etree.write(model.source,
                                  xml_declaration=True,
                                  encoding="utf-8",
                                  standalone=False,
                                  pretty_print=pretty_print_modified_xml_models)
    else:
        model_etree = None

    return model, temp_model_source, model_etree


def resolve_model(model, sed_doc, working_dir):
    """ Resolve the source of a model

    Args:
        model (:obj:`Model`): model whose ``source`` is one of the following

            * A path to a file
            * A URL
            * A MIRIAM URN for an entry in the BioModelsl database (e.g., ``urn:miriam:biomodels.db:BIOMD0000000012``)
            * A reference to another model, using the ``id`` of the other model (e.g., ``#other-model-id``).
              In this case, the model also inherits changes from the parent model.

        sed_doc (:obj:`SedDocument`): parent SED document; used to resolve sources defined by reference to other models
        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)

    Returns:
        :obj:`str`: temporary path to the source of the modified model, if the model needed to be resolved from
    """
    source = model.source

    if source.lower().startswith('urn:'):
        if source.lower().startswith('urn:miriam:biomodels.db:'):
            biomodels_id = source.lower().replace('urn:miriam:biomodels.db:', '')
            url = BIOMODELS_DOWNLOAD_ENDPOINT.format(biomodels_id, biomodels_id)
            response = requests.get(url)
            try:
                response.raise_for_status()
            except Exception:
                raise ValueError('Model `{}` could not be downloaded from BioModels.'.format(biomodels_id))

            temp_file, model.source = tempfile.mkstemp()
            os.close(temp_file)
            with open(model.source, 'wb') as file:
                file.write(response.content)
        else:
            raise NotImplementedError('URN model source `{}` could be resolved.'.format(source))

        return model.source

    elif re.match(r'^http(s)?://', source, re.IGNORECASE):
        response = requests.get(source)
        try:
            response.raise_for_status()
        except Exception:
            raise ValueError('Model could not be downloaded from `{}`.'.format(source))

        temp_file, model.source = tempfile.mkstemp()
        os.close(temp_file)
        with open(model.source, 'wb') as file:
            file.write(response.content)

        return model.source

    elif source.startswith('#'):
        other_model_id = source[1:]
        other_model = next((m for m in sed_doc.models if m.id == other_model_id), None)
        if other_model is None:
            raise ValueError('Relative model source `{}` does not exist.'.format(source))

        model.source = other_model.source
        model.changes = other_model.changes + model.changes
        return resolve_model(model, sed_doc, working_dir)

    else:
        if os.path.isabs(source):
            model.source = source
        else:
            model.source = os.path.join(working_dir, source)

        if not os.path.isfile(model.source):
            raise FileNotFoundError('Model source file `{}` does not exist.'.format(source))

        return None


def apply_changes_to_xml_model(model, model_etree, sed_doc=None, working_dir=None,
                               variable_values=None, range_values=None,
                               validate_unique_xml_targets=True):
    """ Modify an XML-encoded model according to a model change

    Args:
        model (:obj:`Model`): model
        model_etree (:obj:`etree._ElementTree`): element tree for model
        sed_doc (:obj:`SedDocument`, optional): parent SED document; used to resolve sources defined by reference to other models;
            required for compute changes
        working_dir (:obj:`str`, optional): working directory of the SED document (path relative to which models are located);
            required for compute changes
        variable_values (:obj:`dict`, optional): dictionary which contains the value of each variable of each
            compute model change
        range_values (:obj:`dict`, optional): dictionary which contains the value of each range of each
            set value compute model change
        validate_unique_xml_targets (:obj:`bool`, optional): whether to validate the XML targets match
            uniue objects
    """
    for change in model.changes:
        if isinstance(change, ModelAttributeChange):

            # get object to change
            obj_xpath, sep, attr = change.target.rpartition('/@')
            if sep != '/@':
                raise ValueError('target {} is not a valid XPath to an attribute of a model element'.format(change.target))
            objs = eval_xpath(model_etree, obj_xpath, change.target_namespaces)
            if validate_unique_xml_targets and len(objs) != 1:
                raise ValueError('xpath {} must match a single object'.format(obj_xpath))

            ns_prefix, _, attr = attr.rpartition(':')
            if ns_prefix:
                ns = change.target_namespaces.get(ns_prefix, None)
                if ns is None:
                    raise ValueError('No namespace is defined with prefix `{}`'.format(ns_prefix))
                attr = '{{{}}}{}'.format(ns, attr)

            # change value
            for obj in objs:
                obj.set(attr, change.new_value)

        elif isinstance(change, AddElementModelChange):
            parents = eval_xpath(model_etree, change.target, change.target_namespaces)

            if validate_unique_xml_targets and len(parents) != 1:
                raise ValueError('xpath {} must match a single object'.format(change.target))

            try:
                new_elements = etree.fromstring('<root>' + change.new_elements + '</root>').getchildren()
            except etree.XMLSyntaxError as exception:
                raise ValueError('`{}` is invalid XML. {}'.format(change.new_elements, str(exception)))

            for parent in parents:
                for new_element in copy.deepcopy(new_elements):
                    parent.append(new_element)

        elif isinstance(change, ReplaceElementModelChange):
            old_elements = eval_xpath(model_etree, change.target, change.target_namespaces)

            if validate_unique_xml_targets and len(old_elements) != 1:
                raise ValueError('xpath {} must match a single object'.format(change.target))

            try:
                new_elements = etree.parse(io.StringIO('<root>' + change.new_elements + '</root>')).getroot().getchildren()
            except etree.XMLSyntaxError as exception:
                raise ValueError('`{}` is invalid XML. {}'.format(change.new_elements, str(exception)))

            for old_element in old_elements:
                parent = old_element.getparent()

                parent.remove(old_element)

                for new_element in copy.deepcopy(new_elements):
                    parent.append(new_element)

        elif isinstance(change, RemoveElementModelChange):
            elements = eval_xpath(model_etree, change.target, change.target_namespaces)

            if validate_unique_xml_targets and len(elements) != 1:
                raise ValueError('xpath {} must match a single object'.format(change.target))

            for element in elements:
                parent = element.getparent()
                parent.remove(element)

        elif isinstance(change, ComputeModelChange):
            # get the values of model variables referenced by compute model changes
            if variable_values is None:
                model_etrees = {model.id: model_etree}
                iter_variable_values = get_values_of_variable_model_xml_targets_of_model_change(change, sed_doc, model_etrees, working_dir)
            else:
                iter_variable_values = variable_values

            # calculate new value
            new_value = calc_compute_model_change_new_value(change, variable_values=iter_variable_values, range_values=range_values)
            if new_value == int(new_value):
                new_value = str(int(new_value))
            else:
                new_value = str(new_value)

            # get object to change
            obj_xpath, sep, attr = change.target.rpartition('/@')
            if sep != '/@':
                raise ValueError('target {} is not a valid XPath to an attribute of a model element'.format(change.target))
            objs = eval_xpath(model_etree, obj_xpath, change.target_namespaces)
            if validate_unique_xml_targets and len(objs) != 1:
                raise ValueError('xpath {} must match a single object'.format(obj_xpath))

            ns_prefix, _, attr = attr.rpartition(':')
            if ns_prefix:
                ns = change.target_namespaces.get(ns_prefix, None)
                if ns is None:
                    raise ValueError('No namespace is defined with prefix `{}`'.format(ns_prefix))
                attr = '{{{}}}{}'.format(ns, attr)

            # change value
            for obj in objs:
                obj.set(attr, new_value)

        else:
            raise NotImplementedError('Change{} of type {} is not supported.'.format(
                ' ' + change.name if change.name else '', change.__class__.__name__))


def get_values_of_variable_model_xml_targets_of_model_change(change, sed_doc, model_etrees, working_dir):
    """ Get the values of the model variables of a compute model change

    Args:
        change (:obj:`ComputeModelChange`): compute model change
        sed_doc (:obj:`SedDocument`): SED document
        model_etrees (:obj:`dict` of :obj:`str` to :obj:`etree._Element`): map from the ids of models to element
            trees of their sources
        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)

    Returns:
        :obj:`dict`: dictionary which contains the value of each variable of each
            compute model change
    """
    variable_values = {}
    for variable in change.variables:
        variable_model = variable.model
        if variable_model.id not in model_etrees:
            copy_variable_model, temp_model_source, variable_model_etree = resolve_model_and_apply_xml_changes(
                variable_model, sed_doc, working_dir,
                apply_xml_model_changes=True,
                save_to_file=False)
            model_etrees[variable_model.id] = variable_model_etree

            if temp_model_source:
                os.remove(temp_model_source)

        variable_values[variable.id] = get_value_of_variable_model_xml_targets(
            variable, model_etrees)

    return variable_values


def get_value_of_variable_model_xml_targets(variable, model_etrees):
    """ Get the value of a variable of a model

    Args:
        variable (:obj:`Variable`): variable
        model_etrees (:obj:`dict` of :obj:`str` to :obj:`etree._Element`): dictionary that maps the
            ids of models to paths to files which contain their XML definitions

    Returns:
        :obj:`float`: value
    """
    if not variable.target:
        raise NotImplementedError('Compute model change variable `{}` must have a target'.format(variable.id))

    if variable.target.startswith('#'):
        raise NotImplementedError('Variable references to data descriptions are not supported.')

    obj_xpath, sep, attr = variable.target.rpartition('/@')
    if sep != '/@':
        raise ValueError('target {} is not a valid XPath to an attribute of a model element'.format(variable.target))

    et = model_etrees[variable.model.id]
    obj = eval_xpath(et, obj_xpath, variable.target_namespaces)
    if len(obj) != 1:
        raise ValueError('xpath {} must match a single object in model {}'.format(obj_xpath, variable.model.id))

    ns, _, attr = attr.rpartition(':')
    if ns:
        attr = '{{{}}}{}'.format(variable.target_namespaces[ns], attr)

    value = obj[0].get(attr)
    if value is None:
        raise ValueError('Target `{}` is not defined in model `{}`.'.format(variable.target, variable.model.id))
    try:
        value = float(value)
    except ValueError:
        raise ValueError('Target `{}` in model `{}` must be a float.'.format(variable.target, variable.model.id))

    return value


def calc_compute_model_change_new_value(change, variable_values=None, range_values=None):
    """ Calculate the new value of a compute model change

    Args:
        change (:obj:`ComputeModelChange`): change
        variable_values (:obj:`dict`, optional): dictionary which contains the value of each variable of each
            compute model change
        range_values (:obj:`dict`, optional): dictionary which contains the value of each range of each
            set value compute model change

    Returns:
        :obj:`float`: new value
    """
    compiled_math = compile_math(change.math)

    workspace = {}

    if isinstance(change, SetValueComputeModelChange):
        if change.range:
            workspace[change.range.id] = range_values.get(change.range.id, None)
            if workspace[change.range.id] is None:
                raise ValueError('Value of range `{}` is not defined.'.format(change.range.id))

    for param in change.parameters:
        workspace[param.id] = param.value

    for var in change.variables:
        workspace[var.id] = variable_values.get(var.id, None)
        if workspace[var.id] is None:
            raise ValueError('Value of variable `{}` is not defined.'.format(var.id))

    return eval_math(change.math, compiled_math, workspace)


def calc_data_generator_results(data_generator, variable_results):
    """ Calculate the results of a data generator from the results of its variables

    Args:
        data_generator (:obj:`DataGenerator`): data generator
        variable_results (:obj:`VariableResults`): results for the variables of the data generator

    Returns:
        :obj:`numpy.ndarray`: result of data generator
    """
    var_shapes = set()
    max_shape = []
    for var in data_generator.variables:
        var_res = variable_results[var.id]
        var_shape = var_res.shape
        if not var_shape and var_res.size:
            var_shape = (1,)
        var_shapes.add(var_shape)

        max_shape = max_shape + [1 if max_shape else 0] * (var_res.ndim - len(max_shape))
        for i_dim in range(var_res.ndim):
            max_shape[i_dim] = max(max_shape[i_dim], var_res.shape[i_dim])

    if len(var_shapes) > 1:
        warn('Variables for data generator {} do not have consistent shapes'.format(data_generator.id),
             InconsistentVariableShapesWarning)

    compiled_math = compile_math(data_generator.math)

    workspace = {}
    for param in data_generator.parameters:
        workspace[param.id] = param.value

    if not var_shapes:
        value = eval_math(data_generator.math, compiled_math, workspace)
        result = numpy.array(value)

    else:
        for aggregate_func in AGGREGATE_MATH_FUNCTIONS:
            if re.search(aggregate_func + r' *\(', data_generator.math):
                msg = 'Evaluation of aggregate mathematical functions such as `{}` is not supported.'.format(aggregate_func)
                raise NotImplementedError(msg)

        padded_var_shapes = []
        for var in data_generator.variables:
            var_res = variable_results[var.id]
            padded_var_shapes.append(
                list(var_res.shape)
                + [1 if var_res.size else 0] * (len(max_shape) - var_res.ndim)
            )

        result = numpy.full(max_shape, numpy.nan)
        n_dims = result.ndim
        for i_el in range(result.size):
            el_indices = numpy.unravel_index(i_el, result.shape)

            vars_available = True
            for var, padded_shape in zip(data_generator.variables, padded_var_shapes):
                var_res = variable_results[var.id]
                if var_res.ndim == 0:
                    if i_el == 0 and var_res.size:
                        workspace[var.id] = var_res.tolist()
                    else:
                        vars_available = False
                        break

                else:
                    for x, y in zip(padded_shape, el_indices):
                        if (y + 1) > x:
                            vars_available = False
                            break
                    if not vars_available:
                        break

                    workspace[var.id] = var_res[el_indices[0:var_res.ndim]]

            if not vars_available:
                continue

            result_el = eval_math(data_generator.math, compiled_math, workspace)

            if n_dims == 0:
                result = numpy.array(result_el)
            else:
                result.flat[i_el] = result_el

    return result


def calc_data_generators_results(data_generators, variable_results, output, task, make_shapes_consistent=True):
    """ Calculator the values of a list of data generators

    Args:
        data_generators (:obj:`list` of :obj:`DataGenerator`): SED task
        variable_results (:obj:`VariableResults`): results of the SED variables involved in the data generators
        output (:obj:`Output`): SED output
        task (:obj:`Task`): SED task
        make_shapes_consistent (:obj:`bool`, optional): where to make the shapes of the data generators consistent
            (e.g., for concatenation into a table for a report)

    Returns:
        :obj:`tuple`:

            * :obj:`DataGeneratorResults`: values of the data generators
            * :obj:`dict` of :obj:`str` to :obj:`Status`: dictionary that maps the id of each data generator to its status
            * :obj:`Exception`: exception for failures
            * :obj:`bool`: where the task contributes to any of the data generators
    """
    task_contributes_to_data_generators = False
    statuses = {}
    exceptions = []
    results = DataGeneratorResults()

    for data_gen in data_generators:
        vars_available = True
        vars_failed = False
        for variable in data_gen.variables:
            if variable.task == task:
                task_contributes_to_data_generators = True
            if variable.id in variable_results:
                if variable_results.get(variable.id, None) is None:
                    vars_available = False
                    vars_failed = True
            else:
                vars_available = False

        if vars_failed:
            status = Status.FAILED
            msg = 'Data generator {} cannot be calculated because its variables were not successfully produced.'.format(data_gen.id)
            exceptions.append(ValueError(msg))
            result = None

        elif vars_available:
            try:
                result = calc_data_generator_results(data_gen, variable_results)
                status = Status.SUCCEEDED
            except Exception as exception:
                result = None
                exceptions.append(exception)
                status = Status.FAILED

        else:
            status = Status.QUEUED
            result = None

        statuses[data_gen.id] = status
        results[data_gen.id] = result

    if make_shapes_consistent:
        arrays = results.values()
        consistent_arrays = pad_arrays_to_consistent_shapes(arrays)
        for data_gen_id, result in zip(results.keys(), consistent_arrays):
            results[data_gen_id] = result

    if exceptions:
        exception = ValueError('Some generators could not be produced:\n  - {}'.format(
            '\n  '.join(str(exception) for exception in exceptions)))
    else:
        exception = None

    return results, statuses, exception, task_contributes_to_data_generators


def remove_model_changes(sed_doc):
    """ Remove model changes from a SED document

    Args:
        sed_doc (:obj:`SedDocument`): SED document
    """
    for model in sed_doc.models:
        model.changes = []


def remove_algorithm_parameter_changes(sed_doc):
    """ Remove algorithm parameter changes from a SED document

    Args:
        sed_doc (:obj:`SedDocument`): SED document
    """
    for simulation in sed_doc.simulations:
        simulation.algorithm.changes = []


def replace_complex_data_generators_with_generators_for_individual_variables(sed_doc):
    """ Remove model changes from a SED document

    Args:
        sed_doc (:obj:`SedDocument`): SED document
    """
    data_gen_replacements = {}

    for original_data_gen in list(sed_doc.data_generators):
        if len(original_data_gen.parameters) + len(original_data_gen.variables) > 1:
            sed_doc.data_generators.remove(original_data_gen)
            data_gen_replacements[original_data_gen] = []

            for var in original_data_gen.variables:
                new_data_gen = DataGenerator(id='__single_var_gen__' + var.id, variables=[var], math=var.id)
                data_gen_replacements[original_data_gen].append(new_data_gen)
                sed_doc.data_generators.append(new_data_gen)

    for output in sed_doc.outputs:
        if isinstance(output, Report):
            els = 'data_sets'
            props = ['data_generator']

        elif isinstance(output, Plot2D):
            els = 'curves'
            props = ['x_data_generator', 'y_data_generator']

        elif isinstance(output, Plot3D):
            els = 'surfaces'
            props = ['x_data_generator', 'y_data_generator', 'z_data_generator']

        old_els = getattr(output, els)

        new_els = []
        i_single_var_output_el = 0
        for el in old_els:
            replacement_els = [el]
            for prop in props:
                new_replacement_els = []
                for el2 in replacement_els:
                    original_data_gen = getattr(el2, prop)
                    for replaced_data_gen in data_gen_replacements.get(original_data_gen, [original_data_gen]):
                        i_single_var_output_el += 1
                        el3 = copy.copy(el2)
                        el3.id = '__single_var_output_el__' + str(i_single_var_output_el)
                        el3.label = el3.id
                        setattr(el3, prop, replaced_data_gen)
                        new_replacement_els.append(el3)
                replacement_els = new_replacement_els
            new_els.extend(replacement_els)

        setattr(output, els, new_els)


def remove_plots(sed_doc):
    """ Remove plots from a SED document

    Args:
        sed_doc (:obj:`SedDocument`): SED document
    """
    for output in list(sed_doc.outputs):
        if isinstance(output, Plot):
            sed_doc.outputs.remove(output)


def get_first_last_models_executed_by_task(task):
    """ Get the models executed by a task in the order they are executed

    Args:
        task (:obj:`Task`): task

    Returns:
        :obj:`tuple` of :obj:`Model`: models
    """
    if isinstance(task, Task):
        return (task.model, task.model)

    elif isinstance(task, RepeatedTask):
        sub_tasks = sorted(task.sub_tasks, key=lambda sub_task: sub_task.order)
        return (
            get_first_last_models_executed_by_task(sub_tasks[0].task)[0],
            get_first_last_models_executed_by_task(sub_tasks[-1].task)[-1],
        )

    else:
        raise NotImplementedError('Tasks of type `{}` are not supported.'.format(task.__class__.__name__))


def get_models_referenced_by_task(task):
    """ Get the models referenced from a task

    Args:
        task (:obj:`Task`): task

    Returns:
        :obj:`set` of :obj:`Model`: models
    """
    if isinstance(task, Task):
        if task.model:
            return set([task.model])
        else:
            return set()

    elif isinstance(task, RepeatedTask):
        models = set()

        if task.range:
            models.update(get_models_referenced_by_range(task.range))

        for change in task.changes:
            if change:
                models.update(get_models_referenced_by_model_change(change))

        for sub_task in task.sub_tasks:
            if sub_task and sub_task.task:
                models.update(get_models_referenced_by_task(sub_task.task))

        if task.range:
            models.update(get_models_referenced_by_range(task.range))
        for range in task.ranges:
            if range:
                models.update(get_models_referenced_by_range(range))

        return models

    else:
        raise NotImplementedError('Tasks of type `{}` are not supported.'.format(task.__class__.__name__))


def get_models_referenced_by_model_change(change):
    """ Get the models referenced from a model change

    Args:
        change (:obj:`ModelChange`): model change

    Returns:
        :obj:`set` of :obj:`Model`: models
    """
    models = set()

    if isinstance(change, SetValueComputeModelChange):
        if change.model:
            models.add(change.model)

        if change.range:
            models.update(get_models_referenced_by_range(change.range))

        for variable in change.variables:
            if variable.model:
                models.add(variable.model)

    elif isinstance(change, ComputeModelChange):
        for variable in change.variables:
            if variable.model:
                models.add(variable.model)

    return models


def get_models_referenced_by_range(range):
    """ Get the models referenced by a range

    Args:
        range (:obj:`Range`): range

    Returns:
        :obj:`set` of :obj:`Model`: models
    """
    models = set()

    if isinstance(range, FunctionalRange):
        for variable in range.variables:
            if variable.model:
                models.add(variable.model)

        models.update(get_models_referenced_by_range(range.range))

    return models


def get_range_len(range):
    """ Get the length of a range

    Args:
        range (:obj:`Range`): range

    Returns:
        :obj:`int`: length of the range

    Raises:
        :obj:`NotImplementedError`: if range isn't an instance of :obj:`UniformRange`, :obj:`VectorRange`,
            or :obj:`FunctionalRange`.
    """
    if isinstance(range, UniformRange):
        return range.number_of_steps + 1

    elif isinstance(range, VectorRange):
        return len(range.values)

    elif isinstance(range, FunctionalRange):
        return get_range_len(range.range)

    else:
        raise NotImplementedError('Ranges of type `{}` are not supported.'.format(range.__class__.__name__))


def resolve_range(range, model_etrees=None):
    """ Resolve the values of a range

    Args:
        range (:obj:`Range`): range
        model_etrees (:obj:`dict` of :obj:`str` to :obj:`etree._Element`): map from the ids of models to element
            trees of their sources; required to resolve variables of functional ranges

    Returns:
        :obj:`list` of :obj:`float`: values of the range

    Raises:
        :obj:`NotImplementedError`: if range isn't an instance of :obj:`UniformRange`, :obj:`VectorRange`,
            or :obj:`FunctionalRange`.
    """
    if isinstance(range, UniformRange):
        if range.type == UniformRangeType.linear:
            return numpy.linspace(range.start, range.end, range.number_of_steps + 1).tolist()

        elif range.type == UniformRangeType.log:
            return numpy.logspace(numpy.log10(range.start), numpy.log10(range.end), range.number_of_steps + 1).tolist()

        else:
            raise NotImplementedError('UniformRanges of type `{}` are not supported.'.format(range.type.value))

    elif isinstance(range, VectorRange):
        return range.values

    elif isinstance(range, FunctionalRange):
        # compile math
        compiled_math = compile_math(range.math)

        # setup workspace to evaluate math
        workspace = {}

        workspace[range.range.id] = None

        for param in range.parameters:
            workspace[param.id] = param.value

        for var in range.variables:
            if var.symbol:
                raise NotImplementedError('Symbols are not supported for variables of functional ranges')
            if model_etrees[var.model.id] is None:
                raise NotImplementedError('Functional ranges that involve variables of non-XML-encoded models are not supported.')
            workspace[var.id] = get_value_of_variable_model_xml_targets(var, model_etrees)

        # calculate the values of the range
        values = []
        for child_range_value in resolve_range(range.range, model_etrees=model_etrees):
            workspace[range.range.id] = child_range_value

            value = eval_math(range.math, compiled_math, workspace)
            values.append(value)

        # return values
        return values

    else:
        raise NotImplementedError('Ranges of type `{}` are not supported.'.format(range.__class__.__name__))


def get_namespaces_for_sed_object(obj):
    """ Get the namespace prefixes and URIs defined for a SED object

    Args:
        obj (:obj:`libsedml.SedBase`): SED object

    Returns:
        :obj:`dict`: dictionary that maps the prefixes of namespaces to their URIs
    """
    namespaces_obj = obj.getNamespaces()
    namespaces_dict = {}
    for i_namespace in range(namespaces_obj.getNumNamespaces()):
        prefix = namespaces_obj.getPrefix(i_namespace) or None
        uri = namespaces_obj.getURI(i_namespace)
        namespaces_dict[prefix] = uri
    return namespaces_dict


def get_xml_node_namespace_tag_target(etree, target_namespaces=None):
    """ Get the namespace, tag, and target of an XML node

    Args:
        etree (:obj:`etree._Element`): element tree
        target_namespaces (:obj:`dict`, optional): dictionary that maps the prefixes of namespaces to their URIs

    Returns:
        :obj:`tuple`:

            * :obj:`str`: namespace URI
            * :obj:`str`: namespace prefix
            * :obj:`str`: tag
            * :obj:`str`: target for use with SED target XPaths
            * :obj:`dict`: dictionary that maps the prefixes of namespaces to their URIs
    """
    target_namespaces = dict(target_namespaces or {})
    uri, _, tag = etree.tag.rpartition('}')
    if uri:
        uri = uri[1:]
    if uri:
        nsmap_rev = {uri: prefix for prefix, uri in etree.nsmap.items()}
        prefix = nsmap_rev[uri]
        if not prefix:
            target_namespaces_rev = {uri: prefix for prefix, uri in target_namespaces.items()}
            prefix = target_namespaces_rev.get(uri, tag)

        if prefix in target_namespaces:
            if target_namespaces.get(prefix, None) != uri:
                prefix += '2'
                target_namespaces[prefix] = uri
        else:
            target_namespaces[prefix] = uri

        target = prefix + ':' + tag
    else:
        uri = None
        prefix = None
        target = tag

    return (uri, prefix, tag, target, target_namespaces)


def is_model_language_encoded_in_xml(language):
    """ Determine if the model language is encoded in XML

    Args:
        language (:obj:`str`): language

    Returns:
        :obj:`bool`: :obj:`True`, if the model language is encoded in XML
    """
    return (
        language and (
            re.match(ModelLanguagePattern.CellML, language)
            or re.match(ModelLanguagePattern.CopasiML, language)
            or re.match(ModelLanguagePattern.GINML, language)
            or re.match(ModelLanguagePattern.LEMS, language)
            or re.match(ModelLanguagePattern.MorpheusML, language)
            or re.match(ModelLanguagePattern.NeuroML, language)
            or re.match(ModelLanguagePattern.SBML, language)
            or re.match(ModelLanguagePattern.VCML, language)
        )
    )


def does_model_language_use_xpath_variable_targets(language):
    """ Determine if the model language uses XPaths to target variables of data generators

    Args:
        language (:obj:`str`): language

    Returns:
        :obj:`bool`: :obj:`True`, if the model language is encoded in XML
    """
    return (
        re.match(ModelLanguagePattern.CellML, language)
        or re.match(ModelLanguagePattern.CopasiML, language)
        or re.match(ModelLanguagePattern.MorpheusML, language)
        or re.match(ModelLanguagePattern.SBML, language)
        or re.match(ModelLanguagePattern.VCML, language)
    )


def get_all_sed_objects(doc, type=(SedBase, SedIdGroupMixin)):
    """ Get all of the identified objects (instances of :obj:`SedIdGroupMixin`) that
    belong to a SED document

    Args:
        doc (:obj:`SedBase`) SED document
        type (:obj:`type` or :obj:`tuple` of :obj:`type`, optional): type of SED object
            to return
    """
    # get all linked objects
    objs_to_see = [doc]
    seen_objs = [doc]
    while objs_to_see:
        obj = objs_to_see.pop()
        for attr in obj.__dict__.values():
            if isinstance(attr, (SedBase, SedIdGroupMixin)):
                if attr not in seen_objs:
                    objs_to_see.append(attr)
                    seen_objs.append(attr)

            elif isinstance(attr, list):
                for val in attr:
                    if isinstance(val, (SedBase, SedIdGroupMixin)):
                        if val not in seen_objs:
                            objs_to_see.append(val)
                            seen_objs.append(val)

    # filter out elements of a specific type
    return list(filter(lambda obj: isinstance(obj, type), seen_objs))


def get_task_results_shape(task):
    """ Get the shape of the results of a task

    Args:
        task (:obj:`Task`): task

    Returns:
        :obj:`tuple` of :obj:`int`: shape of the results of a task
    """
    if isinstance(task, Task):
        if isinstance(task.simulation, OneStepSimulation):
            return (2,)
        elif isinstance(task.simulation, SteadyStateSimulation):
            return (1,)
        elif isinstance(task.simulation, UniformTimeCourseSimulation):
            return (task.simulation.number_of_steps + 1,)
        else:
            return (None,)

    elif isinstance(task, RepeatedTask):
        shape = [get_range_len(task.range) if task.range else None, len(task.sub_tasks)]
        max_sub_task_shape = []

        for sub_task in task.sub_tasks:
            sub_task_shape = list(get_task_results_shape(sub_task.task))
            max_sub_task_shape = max_sub_task_shape + [0] * max(0, len(sub_task_shape) - len(max_sub_task_shape))
            sub_task_shape = sub_task_shape + [0] * max(0, len(max_sub_task_shape) - len(sub_task_shape))
            max_sub_task_shape = [max(i, j) for i, j in zip(max_sub_task_shape, sub_task_shape)]

        return tuple(shape + max_sub_task_shape)

    return None



""" Utilities for working with BNGL models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-05
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...sedml.data_model import (  # noqa: F401
    SedDocument, ModelAttributeChange, Variable, Symbol,
    Simulation, OneStepSimulation, UniformTimeCourseSimulation,
    Algorithm, AlgorithmParameterChange,
    Task,
    )
from ...utils.core import flatten_nested_list_of_strings
from ...warnings import warn, BioSimulatorsWarning
from .validation import validate_model
import numpy
import os
import re
import types  # noqa: F401

__all__ = ['get_parameters_variables_outputs_for_simulation']


def get_parameters_variables_outputs_for_simulation(model_filename, model_language, simulation_type, algorithm_kisao_id=None,
                                                    change_level=SedDocument, native_ids=False, native_data_types=False):
    """ Get the possible observables for a simulation of a model

    Args:
        model_filename (:obj:`str`): path to model file
        model_language (:obj:`str`): model language (e.g., ``urn:sedml:language:bngl``)
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`, optional): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000019``
            for CVODE)
        change_level (:obj:`types.Type`, optional): level at which model changes will be made (:obj:`SedDocument` or :obj:`Task`)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulations of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
        :obj:`list` of :obj:`Plot`: possible plots of the results of a simulation of the model
    """
    # check model file exists and is valid
    if not isinstance(model_filename, str):
        raise ValueError('`{}` is not a path to a model file.'.format(model_filename))

    if not os.path.isfile(model_filename):
        raise FileNotFoundError('Model file `{}` does not exist.'.format(model_filename))

    errors, _, model = validate_model(model_filename)
    if errors:
        raise ValueError('Model file `{}` is not a valid BNGL or BNGL XML file.\n  {}'.format(
            model_filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  ')))

    if simulation_type not in [OneStepSimulation, UniformTimeCourseSimulation]:
        raise NotImplementedError('`simulation_type` must be `OneStepSimulation` or `UniformTimeCourseSimulation`')

    # parameters
    params = []

    if hasattr(model, 'parameters'):
        for el in model.parameters.items.values():
            escaped_el_name = escape_id(el.name)
            params.append(ModelAttributeChange(
                id=el.name if native_ids else 'value_parameter_{}'.format(escaped_el_name),
                name=None if native_ids else 'Value of parameter "{}"'.format(el.name),
                target='parameters.{}.value'.format(el.name),
                new_value=parse_expression(el.expr, native_data_types=native_data_types),
            ))

    if hasattr(model, 'compartments'):
        for el in model.compartments.items.values():
            params.append(ModelAttributeChange(
                id=el.name if native_ids else 'initial_size_compartment_{}'.format(el.name),
                name=None if native_ids else 'Initial size of {}-D compartment "{}"'.format(el.dim, el.name),
                target='compartments.{}.size'.format(el.name),
                new_value=parse_expression(el.size, native_data_types=native_data_types),
            ))

    if hasattr(model, 'species'):
        for el in model.species.items.values():
            el_pattern = el.pattern.xml['@name'].replace('::', ':')
            escaped_el_pattern = escape_id(el_pattern)
            params.append(ModelAttributeChange(
                id=el_pattern if native_ids else 'initial_amount_species_{}'.format(escaped_el_pattern),
                name=None if native_ids else 'Initial amount of species "{}"'.format(el_pattern),
                target='species.{}.initialCount'.format(el_pattern),
                new_value=parse_expression(el.count, native_data_types=native_data_types),
            ))

    if hasattr(model, 'functions'):
        for el in model.functions.items.values():
            params.append(ModelAttributeChange(
                id=el.name if native_ids else 'expression_function_{}'.format(el.name),
                name=None if native_ids else 'Expression of function "{}({})"'.format(el.name, ', '.join(el.args)),
                target='functions.{}.expression'.format(el.name),
                new_value=str(el.expr),
            ))

    # simulation
    sims = []
    for i_action, action in enumerate(model.actions.items):
        args = action.args
        initial_time = float(args.get('t_start', '0.'))
        output_start_time = initial_time
        output_end_time = args.get('t_end', None)
        output_step_interval = args.get('output_step_interval', None)
        max_sim_steps = args.get('max_sim_steps', None)
        number_of_steps = int(float(args.get('n_steps', args.get('n_output_steps', '1'))))
        seed = args.get('seed', None)
        a_tol = args.get('atol', None)
        r_tol = args.get('rtol', None)
        stop_if = args.get('stop_if', None)

        if action.name == 'simulate':
            method = args.get('method', None)
            if method is None:
                raise ValueError('`simulate` action {} must define a `method` argument.'.format(i_action + 1))
            if method and method[0] == '"':
                method = method[1:-1]

        elif action.name == 'simulate_ode':
            method = 'ode'

        elif action.name == 'simulate_ssa':
            method = 'ssa'

        elif action.name == 'simulate_pla':
            method = 'pla'

        elif action.name == 'simulate_nf':
            method = 'nf'

        elif action.name == 'parameter_scan':
            msg = (
                'Parameter scan action {} was ignored because parameter scan actions are not supported.'
            ).format(i_action + 1)
            warn(msg, BioSimulatorsWarning)
            continue

        elif action.name == 'bifurcate':
            msg = (
                'Bifurcation analysis action {} was ignored because bifurcation analyses actions are not supported.'
            ).format(i_action + 1)
            warn(msg, BioSimulatorsWarning)
            continue

        else:
            continue  # pragma: no cover

        if 'sample_times' in args:
            sample_times_str = args['sample_times'][1:-1].strip()
            if not sample_times_str:
                raise ValueError((
                    'Sample times (`sample_times`) must be a non-empty array of floats '
                    'greater than or equal to the simulation start time (`t_start`).'
                ))
            sample_times = sorted([float(sample_time.strip()) for sample_time in args['sample_times'][1:-1].strip().split(',')])
            output_start_time = sample_times[0]
            output_end_time = sample_times[-1]
            if len(set(numpy.diff(sample_times))) <= 1:
                number_of_steps = len(sample_times) - 1
            else:
                msg = (
                    'Non-uniformly-distributed sample times (`sample_times`) for action {} were ignored '
                    'because they cannot be translated into SED-ML.'
                ).format(i_action + 1)
                warn(msg, BioSimulatorsWarning)
        if output_end_time is None:
            raise ValueError('`Simulation end time (`t_end`) must be set for `{}` action {}.'.format(action.name, i_action + 1))

        if method == 'ode':
            algorithm_kisao_id = 'KISAO_0000019'
        elif method == 'ssa':
            algorithm_kisao_id = 'KISAO_0000029'
        elif method == 'pla':
            algorithm_kisao_id = 'KISAO_0000524'
        elif method == 'nf':
            algorithm_kisao_id = 'KISAO_0000263'

        sim = UniformTimeCourseSimulation(
            id='simulation_{}'.format(i_action),
            initial_time=initial_time,
            output_start_time=output_start_time,
            output_end_time=float(output_end_time),
            number_of_steps=number_of_steps,
            algorithm=Algorithm(
                kisao_id=algorithm_kisao_id,
            )
        )

        if seed is not None:
            sim.algorithm.changes.append(AlgorithmParameterChange(
                kisao_id='KISAO_0000488', new_value=int(float(seed)) if native_data_types else str(seed)))
        if a_tol is not None:
            sim.algorithm.changes.append(AlgorithmParameterChange(
                kisao_id='KISAO_0000211', new_value=float(a_tol) if native_data_types else str(a_tol)))
        if r_tol is not None:
            sim.algorithm.changes.append(AlgorithmParameterChange(
                kisao_id='KISAO_0000209', new_value=float(r_tol) if native_data_types else str(r_tol)))
        if output_step_interval is not None:
            msg = (
                'Output step interval (`output_step_interval`) was ignored for action {} because this cannot be '
                'encoded into a SED-ML uniform time course.'
            ).format(i_action + 1)
            warn(msg, BioSimulatorsWarning)
        if max_sim_steps is not None:
            sim.algorithm.changes.append(AlgorithmParameterChange(
                kisao_id='KISAO_0000415', new_value=int(float(max_sim_steps)) if native_data_types else str(max_sim_steps)))
        if stop_if is not None:
            if stop_if[0] == '"' and stop_if[-1] == '"':
                stop_if = stop_if[1:-1]
            sim.algorithm.changes.append(AlgorithmParameterChange(
                kisao_id='KISAO_0000525', new_value=stop_if))

        sims.append(sim)

    if len(sims) == 0:
        if simulation_type == OneStepSimulation:
            sims.append(
                OneStepSimulation(
                    id='simulation',
                    step=1.,
                    algorithm=Algorithm(
                        kisao_id=algorithm_kisao_id or 'KISAO_0000019',
                    )
                )
            )
        else:
            sims.append(
                UniformTimeCourseSimulation(
                    id='simulation',
                    initial_time=0.,
                    output_start_time=0.,
                    output_end_time=1.,
                    number_of_steps=10,
                    algorithm=Algorithm(
                        kisao_id=algorithm_kisao_id or 'KISAO_0000019',
                    )
                )
            )

    # observables
    vars = []

    vars.append(Variable(
        id=None if native_ids else 'time',
        name=None if native_ids else 'Time',
        symbol=Symbol.time,
    ))

    if hasattr(model, 'molecule_types'):
        for el in model.molecule_types.items.values():
            el_molecule = str(el)
            escaped_el_molecule = escape_id(el_molecule)

            multiple_states = False
            for component in el.molecule.components:
                if len(component.states) > 1:
                    multiple_states = True
                    break

            if not multiple_states:
                vars.append(Variable(
                    id=el_molecule if native_ids else 'amount_molecule_{}'.format(escaped_el_molecule),
                    name=None if native_ids else 'Dynamics of molecule "{}"'.format(el_molecule),
                    target='molecules.{}.count'.format(el_molecule),
                ))

    if hasattr(model, 'species'):
        for el in model.species.items.values():
            el_pattern = el.pattern.xml['@name'].replace('::', ':')
            escaped_el_pattern = escape_id(el_pattern)
            vars.append(Variable(
                id=el_pattern if native_ids else 'amount_species_{}'.format(escaped_el_pattern),
                name=None if native_ids else 'Dynamics of species "{}"'.format(el_pattern),
                target='species.{}.count'.format(el_pattern),
            ))

    if hasattr(model, 'observables'):
        molecule_types = list(str(el.molecule) for el in model.molecule_types.items.values())
        for el in model.observables.items.values():
            if el.type == 'Molecules':
                for pattern in el.patterns:
                    pattern = str(pattern)
                    if pattern not in molecule_types:
                        escaped_pattern = escape_id(pattern)
                        vars.append(Variable(
                            id=pattern if native_ids else 'amount_molecule_{}'.format(escaped_pattern),
                            name=None if native_ids else 'Dynamics of molecule "{}"'.format(pattern),
                            target='molecules.{}.count'.format(pattern),
                        ))

    return (params, sims, vars, [])


def escape_id(id):
    return re.sub(r'[^a-zA-Z0-9_]', '_', id)


def parse_expression(value, native_data_types=False):
    """ Optionally parse an expression

    Args:
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types

    Returns:
        :obj:`object`: expression or parsed expression
    """
    if native_data_types:
        try:
            return float(value)
        except ValueError:
            return str(value)
    else:
        return str(value)



""" Tool for managing licenses through environment variables

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-09-1
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import abc
import contextlib
import os


class LicenseManager(contextlib.AbstractContextManager):
    """ Base singleton class for managing setting up licenses for software packages from environment variables """

    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(LicenseManager, cls).__new__(cls)
        return cls.__instance

    @property
    @classmethod
    @abc.abstractmethod
    def ENV_VAR_PREFIX(cls):
        """ Get the prefix for environment variables

        Returns:
            :obj:`str`: prefix for environment variables
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def is_package_available(self):
        """ Determine whether the package is installed and available

        Returns:
            :obj:`bool`: whether the package is installed and available
        """
        pass  # pragma: no cover

    def get_keys_from_env_vars(self):
        """ Get the license keys for the software package from environment variables

        Returns:
            :obj:`dict`: environment license variables for a software package
        """
        keys = {}
        for key, val in os.environ.items():
            if key.startswith(self.ENV_VAR_PREFIX + '_') and len(key) > len(self.ENV_VAR_PREFIX) + 1:
                key = key[len(self.ENV_VAR_PREFIX) + 1:]
                keys[key] = val
        return keys

    @abc.abstractmethod
    def start(self):
        """ Initialize the usage of the software package with license keys """
        pass  # pragma: no cover

    @abc.abstractmethod
    def end(self):
        """ Terminate usage of the software package with license keys """
        pass  # pragma: no cover

    def __enter__(self):
        """ Enter a context """
        self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        """ Exit a context

        Args:
            exc_type
            exc_value
            traceback
        """
        self.end()



""" Methods for generating SED plots

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-23
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..sedml.data_model import Plot2D, Plot3D  # noqa: F401
from ..report.data_model import DataGeneratorResults  # noqa: F401
from ..warnings import warn
from .data_model import VizFormat
from .warnings import IllogicalVizWarning
from matplotlib import cm as ColorMap
from matplotlib import pyplot
import matplotlib  # noqa: F401
import numpy
import os


__all__ = ['write_plot_2d', 'write_plot_3d']


def write_plot_2d(plot, data_generator_results, base_path, rel_path, format=VizFormat.pdf, style='seaborn-paper'):
    """ Generate a 2D plot

    Args:
        plot (:obj:`Plot2D`): description of plot
        data_generator_results (:obj:`DataGeneratorResults`): results of data generators
        base_path (:obj:`str`): base path to save plot
        rel_path (:obj:`str`): path to save results relative to :obj:`base_path`
        format (:obj:`VizFormat`, optional): format
        style (:obj:`str`, optional): matplotlib style
    """
    figure = pyplot.figure()
    axes = figure.gca()

    x_names = set()
    y_names = set()

    x_scales = set()
    y_scales = set()

    curve_names = []
    for curve in plot.curves:
        x_id = curve.x_data_generator.id
        y_id = curve.y_data_generator.id

        x_result = data_generator_results[x_id]
        y_result = data_generator_results[y_id]

        with pyplot.style.context(style):
            if x_result.ndim > 1:
                x_result = x_result.reshape((numpy.prod(x_result.shape[0:-1]), x_result.shape[-1])).transpose()
            if y_result.ndim > 1:
                y_result = y_result.reshape((numpy.prod(y_result.shape[0:-1]), y_result.shape[-1])).transpose()

            axes.plot(x_result, y_result)

        x_name = curve.x_data_generator.name or x_id
        y_name = curve.y_data_generator.name or y_id
        x_names.add(x_name)
        y_names.add(y_name)

        x_scales.add(curve.x_scale)
        y_scales.add(curve.y_scale)

        curve_names.append(curve.name or curve.id)

    with pyplot.style.context(style):
        if len(x_names) == 1:
            axes.set_xlabel(list(x_names)[0])
        else:
            warn('A title could not be inferred for the X axis because the X data generators have inconsistent names.',
                 IllogicalVizWarning)
            axes.set_xlabel('X')
        if len(y_names) == 1:
            axes.set_ylabel(list(y_names)[0])
        else:
            warn('A title could not be inferred for the Y axis because the Y data generators have inconsistent names.',
                 IllogicalVizWarning)
            axes.set_ylabel('Y')

        if len(x_scales) == 1:
            axes.set_xscale(list(x_scales)[0].value)
        else:
            warn('Curves have inconsistent x axis scales. All curves will be plotted in linear scale.',
                 IllogicalVizWarning)
        if len(y_scales) == 1:
            axes.set_yscale(list(y_scales)[0].value)
        else:
            warn('Curves have inconsistent y axis scales. All curves will be plotted in linear scale.',
                 IllogicalVizWarning)

        if len(plot.curves) > 1:
            axes.legend(curve_names)

    dirname = os.path.dirname(os.path.join(os.path.join(base_path, rel_path)))
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    figure.savefig(os.path.join(base_path, rel_path + '.' + format.value))


def write_plot_3d(plot, data_generator_results, base_path, rel_path, format=VizFormat.pdf,
                  style='seaborn-paper',
                  colormaps=[ColorMap.viridis, ColorMap.plasma, ColorMap.inferno, ColorMap.magma, ColorMap.cividis]):
    """ Generate a 3D plot

    Args:
        plot (:obj:`Plot3D`): description of plot
        data_generator_results (:obj:`DataGeneratorResults`): results of data generators
        base_path (:obj:`str`): base path to save plot
        rel_path (:obj:`str`): path to save results relative to :obj:`base_path`
        format (:obj:`VizFormat`, optional): format
        style (:obj:`str`, optional): matplotlib style
        colormaps (:obj:`list` of :obj:`matplotlib.colors.LinearSegmentedColormap`, optional): colormaps
    """
    figure = pyplot.figure()
    axes = figure.gca(projection='3d')

    x_names = set()
    y_names = set()
    z_names = set()

    x_scales = set()
    y_scales = set()
    z_scales = set()

    plotted_surfaces = []
    surface_names = []
    for i_surface, surface in enumerate(plot.surfaces):
        x_id = surface.x_data_generator.id
        y_id = surface.y_data_generator.id
        z_id = surface.z_data_generator.id

        x_result = data_generator_results[x_id]
        y_result = data_generator_results[y_id]
        z_result = data_generator_results[z_id]

        n_dim = 0
        extra_dims_lens = set()

        if x_result.ndim > 2 and numpy.any(numpy.array(x_result.shape[0:-2]) != 1):
            n_dim = max(n_dim, x_result.ndim)
            extra_dims_lens.add(numpy.prod(x_result.shape[0:-2]))
            x_result = x_result.reshape((numpy.prod(x_result.shape[0:-2]), x_result.shape[-2], x_result.shape[-1]))
        if y_result.ndim > 2 and numpy.any(numpy.array(y_result.shape[0:-2]) != 1):
            n_dim = max(n_dim, y_result.ndim)
            extra_dims_lens.add(numpy.prod(y_result.shape[0:-2]))
            y_result = y_result.reshape((numpy.prod(y_result.shape[0:-2]), y_result.shape[-2], y_result.shape[-1]))
        if z_result.ndim > 2 and numpy.any(numpy.array(z_result.shape[0:-2]) != 1):
            n_dim = max(n_dim, z_result.ndim)
            extra_dims_lens.add(numpy.prod(z_result.shape[0:-2]))
            z_result = z_result.reshape((numpy.prod(z_result.shape[0:-2]), z_result.shape[-2], z_result.shape[-1]))

        if extra_dims_lens:
            warn('3D surface plots of {}-dimensional data may not be informative.'.format(n_dim), IllogicalVizWarning)

        if len(extra_dims_lens) > 1:
            raise ValueError('Multidimensional data must have consistent shapes.')
        elif extra_dims_lens:
            n_surfaces = list(extra_dims_lens)[0]
        else:
            n_surfaces = 1

        for j_surface in range(n_surfaces):
            if x_result.ndim <= 2:
                x = x_result
            elif x_result.shape[0] == 1:
                x = x_result.squeeze(axis=0)
            else:
                x = x_result[j_surface, :, :]

            if y_result.ndim <= 2:
                y = y_result
            elif y_result.shape[0] == 1:
                y = y_result.squeeze(axis=0)
            else:
                y = y_result[j_surface, :, :]

            if z_result.ndim <= 2:
                z = z_result
            elif z_result.shape[0] == 1:
                z = z_result.squeeze(axis=0)
            else:
                z = z_result[j_surface, :, :]

            with pyplot.style.context(style):
                plotted_surfaces.append(axes.plot_surface(
                    x, y, z, cmap=colormaps[((i_surface * len(plot.surfaces)) + j_surface) % len(colormaps)]))

        x_name = surface.x_data_generator.name or x_id
        y_name = surface.y_data_generator.name or y_id
        z_name = surface.z_data_generator.name or z_id
        x_names.add(x_name)
        y_names.add(y_name)
        z_names.add(z_name)

        x_scales.add(surface.x_scale)
        y_scales.add(surface.y_scale)
        z_scales.add(surface.z_scale)

        surface_names.append(surface.name or surface.id)

    with pyplot.style.context(style):
        if len(x_names) == 1:
            axes.set_xlabel(list(x_names)[0])
        else:
            warn('A title could not be inferred for the X axis because the X data generators have inconsistent names.',
                 IllogicalVizWarning)
            axes.set_xlabel('X')
        if len(y_names) == 1:
            axes.set_ylabel(list(y_names)[0])
        else:
            warn('A title could not be inferred for the Y axis because the Y data generators have inconsistent names.',
                 IllogicalVizWarning)
            axes.set_ylabel('Y')
        if len(z_names) == 1:
            axes.set_zlabel(list(z_names)[0])
        else:
            warn('A title could not be inferred for the Z axis because the Z data generators have inconsistent names.',
                 IllogicalVizWarning)
            axes.set_zlabel('Z')

        if len(x_scales) == 1:
            axes.set_xscale(list(x_scales)[0].value)
        else:
            warn('Curves have inconsistent x axis scales. All surfaces will be plotted in linear scale.',
                 IllogicalVizWarning)
        if len(y_scales) == 1:
            axes.set_yscale(list(y_scales)[0].value)
        else:
            warn('Curves have inconsistent y axis scales. All surfaces will be plotted in linear scale.',
                 IllogicalVizWarning)
        if len(z_scales) == 1:
            axes.set_zscale(list(z_scales)[0].value)
        else:
            warn('Curves have inconsistent z axis scales. All surfaces will be plotted in linear scale.',
                 IllogicalVizWarning)

        for plotted_surface, surface_name in zip(plotted_surfaces, surface_names):
            colorbar = figure.colorbar(plotted_surface)
            colorbar.ax.set_title(surface_name)

    dirname = os.path.dirname(os.path.join(os.path.join(base_path, rel_path)))
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    figure.savefig(os.path.join(base_path, rel_path + '.' + format.value))



""" Data model for references (e.g., journal articles, books)

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-09-05
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import abc
import dataclasses
import typing

__all__ = ['Reference', 'JournalArticle', 'PubMedCentralOpenAccesGraphic']


class Reference(abc.ABC):
    @abc.abstractmethod
    def get_citation(self):
        """ Get a human-readable citation for a reference

        Returns:
            :obj:`str`: human-readable citation
        """
        pass  # pragma: no cover


@dataclasses.dataclass
class JournalArticle(Reference):
    """ Journal article

    Attributes:
        pubmed_id (:obj:`str`): PubMed id
        pubmed_central_id (:obj:`str`): PubMed Central id
        doi (:obj:`str`): DOI
        authors (:obj:`list` of :obj:`str`): authors
        title (:obj:`str`): title
        journal (:obj:`str`): journal
        volume (:obj:`str`): volume
        issue (:obj:`str`): issue
        pages (:obj:`str`): pages
        year (:obj:`str`): year
        date (:obj:`str`): publication date
    """
    pubmed_id: str = None
    pubmed_central_id: str = None
    doi: str = None
    authors: typing.List[str] = dataclasses.field(default_factory=lambda: [])
    title: str = None
    journal: str = None
    volume: str = None
    issue: str = None
    pages: str = None
    year: int = None
    date: str = None

    def get_citation(self):
        """ Format a citation for a reference (e.g., "Authors. Title. Journal volume, issue: pages (year)".).

        Returns:
            :obj:`str`: formatted citation for a reference
        """

        if len(self.authors) > 1:
            authors = '{} & {}'.format(', '.join(self.authors[0:-1]), self.authors[-1])
        elif self.authors:
            authors = self.authors[0]
        else:
            authors = None

        citation = []

        if authors:
            citation.append(authors)

        if self.title:
            if citation:
                citation.append('. ')
            citation.append(self.title)
            if self.title[-1] in '.?':
                sep = ' '
            else:
                sep = '. '
        else:
            sep = ''

        if self.journal:
            citation.append(sep)
            citation.append(self.journal)

        if self.volume is not None:
            if citation:
                citation.append(' ')
            citation.append(self.volume)

        if self.issue:
            if citation:
                citation.append(', ')
            citation.append(self.issue)

        if self.pages:
            if citation:
                citation.append(': ')
            citation.append(self.pages)

        if self.year:
            if citation:
                citation.append(' ')
            citation.append('(' + str(self.year) + ')')

        citation = ''.join(citation)
        if citation:
            if citation[-1] in '.?':
                return citation
            else:
                return citation + '.'
        else:
            return None


@dataclasses.dataclass
class PubMedCentralOpenAccesGraphic(object):
    """ A PubMed Central open access graphic

    Attributes:
        id (:obj:`str`): PubMed Central id
        label (:obj:`str`): label (e.g., ``Figure 1``)
        caption (:obj:`str`): caption
        filename (:obj:`str`): local path to JPEG file for graphic
    """
    id: str = None
    label: str = None
    caption: str = None
    filename: str = None



""" Utilities for working with Identifiers.org

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-08-04
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_app_dirs
import dataclasses
import datetime
import dateutil.parser
import functools
import os.path
import re
import requests_cache
import typing

__all__ = [
    'IdentifiersOrgCountry',
    'IdentifiersOrgInstitution',
    'IdentifiersOrgNamespaceResource',
    'IdentifiersOrgNamespace',
    'InvalidIdentifiersOrgUri',
    'get_identifiers_org_namespaces',
    'get_identifiers_org_namespace',
    'validate_identifiers_org_uri',
]


NAMESPACES_ENDPOINT = 'https://registry.api.identifiers.org/resolutionApi/getResolverDataset'


@dataclasses.dataclass
class IdentifiersOrgCountry(object):
    code: str
    name: str


@dataclasses.dataclass
class IdentifiersOrgInstitution(object):
    id: int
    name: str
    home_url: str
    description: str
    ror_id: str
    country: IdentifiersOrgCountry


@dataclasses.dataclass
class IdentifiersOrgNamespaceResource(object):
    id: int
    mir_id: str
    name: str
    description: str
    url_pattern: str
    official: bool
    provider_code: str
    sample_id: str
    home_url: str
    institution: IdentifiersOrgInstitution
    country: IdentifiersOrgCountry
    deprecated: bool
    deprecated_date: datetime.datetime


@dataclasses.dataclass
class IdentifiersOrgNamespace(object):
    id: int
    mir_id: str
    prefix: str
    name: str
    description: str
    pattern: re.Pattern
    embedded_in_lui: bool
    sample_id: str
    resources: typing.List[IdentifiersOrgNamespaceResource]
    deprecated: bool
    deprecated_date: datetime.datetime
    created: datetime.datetime
    modified: datetime.datetime


class InvalidIdentifiersOrgUri(Exception):
    """ An invalid Identifers.org URI """
    pass


@functools.lru_cache(maxsize=None)
def get_identifiers_org_namespaces():
    """ Get the namespaces registered with `Identifiers.org <https://identifiers.org/>`_.

    Returns:
        :obj:`dict`: dictionary that maps the prefix of each Identifiers.org namespace or
            a tuple of its provider code and its prefix to its attributes
    """
    filename = os.path.join(get_app_dirs().user_cache_dir, 'identifiers-org')
    session = requests_cache.CachedSession(filename, expire_after=7*24*60*60)

    response = session.get(NAMESPACES_ENDPOINT)
    response.raise_for_status()

    namespaces = {}
    for namespace in response.json()['payload']['namespaces']:
        resources = []
        for resource in namespace['resources']:
            resources.append(IdentifiersOrgNamespaceResource(
                id=resource['id'],
                mir_id=resource['mirId'],
                name=resource['name'],
                description=resource['description'],
                url_pattern=resource['urlPattern'],
                official=resource['official'],
                provider_code=resource['providerCode'],
                sample_id=resource['sampleId'],
                home_url=resource['resourceHomeUrl'],
                institution=IdentifiersOrgInstitution(
                    id=resource['institution']['id'],
                    name=resource['institution']['name'],
                    home_url=resource['institution']['homeUrl'],
                    description=resource['institution']['description'],
                    ror_id=resource['institution']['rorId'],
                    country=IdentifiersOrgCountry(
                        code=resource['institution']['location']['countryCode'],
                        name=resource['institution']['location']['countryName'],
                    ),
                ),
                country=IdentifiersOrgCountry(
                    code=resource['location']['countryCode'],
                    name=resource['location']['countryName'],
                ),
                deprecated=resource['deprecated'],
                deprecated_date=dateutil.parser.parse(resource['deprecationDate']) if resource['deprecationDate'] else None,
            ))

        namespace_obj = IdentifiersOrgNamespace(
            id=namespace['id'],
            mir_id=namespace['mirId'],
            prefix=namespace['prefix'],
            name=namespace['name'],
            description=namespace['description'],
            pattern=re.compile(namespace['pattern']),
            embedded_in_lui=namespace['namespaceEmbeddedInLui'],
            sample_id=namespace['sampleId'],
            resources=resources,
            deprecated=namespace['deprecated'],
            deprecated_date=dateutil.parser.parse(namespace['deprecationDate']) if namespace['deprecationDate'] else None,
            created=dateutil.parser.parse(namespace['created']),
            modified=dateutil.parser.parse(namespace['modified']),
        )
        namespaces[namespace['prefix'].lower()] = namespace_obj
        for resource in namespace['resources']:
            namespaces[resource['providerCode'].lower() + '/' + namespace['prefix'].lower()] = namespace_obj

    return namespaces


def get_identifiers_org_namespace(prefix):
    """ Get an Identifiers.org namespace

    Args:
        prefix (:obj:`str`): prefix (e.g., ``biosimulators``) or ``/``-separated tuple of a
            provider code and prefix (e.g, ``icahn/biosimulators``)

    Returns:
        :obj:`IdentifiersOrgNamespace`: namespace

    Raises:
        :obj:`InvalidIdentifiersOrgUri`: if the prefix is not defined
    """
    try:
        return get_identifiers_org_namespaces()[prefix]
    except KeyError:
        raise InvalidIdentifiersOrgUri('`{}` is not a valid prefix of a Identifiers.org namespace.'.format(prefix))


def validate_identifiers_org_uri(uri):
    """ Determine whether a URI is a validate for one of the namespaces registered with Identifiers.org

    Raises:
        :obj:`InvalidIdentifiersOrgUri`: if the URI is not a valid with one of the namespaces registered with Identifiers.org
    """
    match = re.match(r'^https?://identifiers\.org/((([^/:]+)/([^/:]+)|[^/:]+)([/:])(.+))$', uri)
    try:
        namespace = get_identifiers_org_namespace(match.group(2).lower())
    except InvalidIdentifiersOrgUri:
        if match.group(3):
            namespace = get_identifiers_org_namespace(match.group(3).lower())
        else:
            raise

    if (
        not (
            not (match.group(5) in [':', '/'] and ':' in match.group(6)) and
            re.match(namespace.pattern, match.group(6))
        ) and
        not re.match(namespace.pattern, match.group(1)) and
        not (match.group(4) and re.match(namespace.pattern, match.group(4) + match.group(5) + match.group(6)))
    ):
        raise InvalidIdentifiersOrgUri('Identifier `{}` is not valid for the `{}` namespace.'.format(
            match.group(4) + match.group(5) + match.group(6) if match.group(4) else match.group(6),
            namespace.prefix))
